(*****************************************************************************
 * Lexical Analyzer for Tony Language
 * Authors: Giannis Tsiouris, Chris Stavrakakis
 *
 * 
 * This file is part of Tonyc. 
 *
 * Tonyc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*******************************************************************************)

(* Header Section *)
{   
  open Parser
  open Lexing 

  (* Update line number for proper error positioning *)
  let count_line lexbuf = 
      let pos = lexbuf.lex_curr_p in
      lexbuf.lex_curr_p <- {pos with pos_lnum=pos.pos_lnum + 1;
        pos_bol = pos.pos_cnum}

let create_hashtable size init =
    let tbl = Hashtbl.create size in
    List.iter (fun (key, data) -> Hashtbl.add tbl key data) init;
     tbl

let escape_table = 
  let tbl = Hashtbl.create 8 in
  List.iter (fun (key, data) ->
    Hashtbl.add tbl key data) [('n',10);('t',9);('r',13);('0',0);('\\',92);('\'',39);('\"',34);(' ',32)]; tbl
}	



(* Definitions Section *)
let D = ['0'-'9']               (* Digits *)
let L = ['A'-'Z' 'a'-'z']       (* Letters *)
let W = [' ' '\t' '\n' '\r']    (* White Spaces *)
let HEX = (D | ['a'-'f' 'A'-'F']) (* Hex number *)
let XNN = "\\x" HEX HEX
let ESC = 'n'|'r'|'0'|'t'|'\\'|'\''|'\"'




(* Rules Section *)
rule scanner = parse
    | "<*"      {(*print_endline "comments start";*) comments 0 lexbuf}
    | "and"     { T_and }
    | "bool"    { T_bool }
    | "char"    { T_char }
    | "decl"    { T_decl }
    | "def"     { T_def }
    | "else"    { T_else }
    | "elsif"   { T_elsif }
    | "end"     { T_end }
    | "exit"    { T_exit }
    | "false"   { T_false }
    | "for"     { T_for }
    | "head"    { T_head }
    | "if"      { T_if }
    | "int"     { T_int }
    | "list"    { T_list }
    | "mod"     { T_mod }
    | "new"     { T_new }
    | "nil"     { T_nil }
    | "nil?"    { T_nilQ }
    | "not"     { T_not } 
    | "or"      { T_or }
    | "ref"     { T_ref }
    | "return"  { T_return }
    | "skip"    { T_skip }
    | "tail"    { T_tail }
    | "true"    { T_true }
    (* Operators *)
    | '+'       { T_plus }
    | '-'       { T_minus }
    | '*'       { T_mul }
    | '/'       { T_div }
    | '#'       { T_ins }
    | '='       { T_eq }
    | "<>"      { T_neq }
    | '<'       { T_less }
    | '>'       { T_more }
    | "<="      { T_leq }
    | ">="      { T_meq }
    (* Separators *)
    | '('       { T_lpar }
    | ')'       { T_rpar }
    | '['       { T_lbra }
    | ']'       { T_rbra }
    | ','       { T_comma }
    | ';'       { T_colon }
    | ':'       { T_dots }
    | ":="      { T_assign } 
    (* Identifier *)
    | L(L|D|"_"|'?')* as word           { T_id word}                    
    (* Integer Constant *) 
    | D+ as no                          { T_const (int_of_string no)}         
    (* Single Character *) 
    | "'" ([^'\\' '"' '\'']) "'"        {T_letter (Lexing.lexeme_char lexbuf 1)}
    | '\'' ('\\' ESC | ' ')'\''         { let ch = char_of_int 
       (Hashtbl.find escape_table (Lexing.lexeme_char lexbuf 2))  in T_letter ch} 
    | '\'' XNN '\'' as carac {
      let c_h = String.sub carac 1 4  in c_h.[0]<-'0'; 
      let chara = char_of_int (int_of_string c_h) in T_letter chara }
    | '\n'                              { count_line lexbuf; scanner lexbuf }
    | (W # ['\n'])+                     { scanner lexbuf }
    | '"' ( [^'"' '\n'] | '\\' ESC )+ '"' as str    { T_string str}
    | "%" [^ '\n']+                     { scanner lexbuf}   (* In-Line Comment *)
    | eof                               { T_eof }
    | _ as err                       { 
        Printf.eprintf "Invalid character: '%c' (ascii: %d) in line %d \n"
        err (Char.code err) (lexbuf.lex_curr_p.pos_lnum);
        scanner lexbuf }
and comments level = parse
    | "*>"  { (*Printf.printf "comments (%d) end\n" level;*)
              if level = 0 then scanner lexbuf
              else comments (level-1) lexbuf }
    | "<*"  { (*Printf.printf "comments (%d) start\n" (level+1);*)
              comments (level+1) lexbuf }
    | '\n'  { count_line lexbuf; comments level lexbuf }
    | _     { comments level lexbuf }
    | eof   { print_endline "comments are not closed" ; T_eof}
