%{
(******************************************************************************
 * Parser for Tony Language
 * Authors: Giannis Tsiouris, Chris Stavrakakis
 *
 * 
 * This file is part of Tonyc. 
 *
 * Tonyc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
******************************************************************************)

(* set et, ts=2, sw=2 *)
open Format
open Lexing
open Printf

open Identifier
open Symbol
open Types
open Error

(* Declare module synonyms *)
module U = Utils        (* For scope-testing and pretty printing *)
module I = Intermediate (* Auxilary functions and datatypes for Intermediate
                         * code *)
open Intermediate


(* API for Function Management
 * We need a Function stack:
 *   to deal with function return values,
 *   to check if every function has got a (at least one!) return stmt,
 *   to get unique function id for asm printing. *)
let (func_stack:(Symbol.entry * int) Stack.t) = Stack.create ()

(* Semantic Value Struct 
 * This struct is used as a return value by expressions (non-terminal symbols)
 * and is used for the Semantic Analysis of a Tony program. Type sem_val
 * consists of information about the expression (value_type and expr_type) and
 * the Intermediate code to be generated (PLACE, TRUE, FALSE, NEXT lists). *)
type value_type = Lval | Rval

type sem_val = {
  value : value_type;
  expr_type : Types.typ;
  mutable place : I.operand_t;
  mutable l_next : int list;
  mutable l_true : int list;
  mutable l_false : int list;
}

(* Default Semantic Value for initialization *)
let def_sem = {
  value = Rval;
  expr_type = Types.TYPE_none;
  place = I.Invalid;
  l_next = [];
  l_true = [];
  l_false= [];
}

(* Initialize Symbol Table and other initializations *)
let () = 
  initSymbolTable 256

(* External functions that make up Tony's standard library. *)
let library_functions = [
  (* Input/Output *)
  "puti", TYPE_proc, ["n", TYPE_int, PASS_BY_VALUE];
  "putb", TYPE_proc, ["b", TYPE_bool, PASS_BY_VALUE];
  "putc", TYPE_proc, ["c", TYPE_char, PASS_BY_VALUE];
  "puts", TYPE_proc, ["s", TYPE_array (TYPE_char, 0), PASS_BY_VALUE];
  "geti", TYPE_int, [];
  "getb", TYPE_bool, [];
  "getc", TYPE_char, [];
  "gets", TYPE_proc, [("n", TYPE_int, PASS_BY_VALUE);
      ("s", TYPE_array (TYPE_char, 0), PASS_BY_VALUE)];
  (* Convertion *)
  "abs", TYPE_int, ["n", TYPE_int, PASS_BY_VALUE];
  "ord", TYPE_int, ["c", TYPE_char, PASS_BY_VALUE];
  "chr", TYPE_char, ["n", TYPE_int, PASS_BY_VALUE];
  (* String Manipulation *)
  "strlen", TYPE_int, ["s", TYPE_array (TYPE_char, 0), PASS_BY_VALUE];
  "strcmp", TYPE_int, [("s1", TYPE_array (TYPE_char, 0), PASS_BY_VALUE);
                       ("s2", TYPE_array (TYPE_char, 0), PASS_BY_VALUE)];
  "strcpy", TYPE_proc, [("s1", TYPE_array (TYPE_char, 0), PASS_BY_VALUE);
                        ("s2", TYPE_array (TYPE_char, 0), PASS_BY_VALUE)];
  "strcat", TYPE_proc, [("s1", TYPE_array (TYPE_char, 0), PASS_BY_VALUE);
                        ("s2", TYPE_array (TYPE_char, 0), PASS_BY_VALUE)];
]

(* Function that inserts Standard Library in Main's Scope *)
let create_func (id, typ, args) =
  let f = newFunction (id_make id) true 
  in
    openScope ();
    List.iter (fun (name, typeee, mode) -> ignore (newParameter (id_make name)
      typeee mode f true)) args;
    endFunctionHeader f typ;
    closeScope ();
    (* HACK: reset uniq id counter because builtins use other naming
     * convention from user-defined functions (they have uniq_id = 0). *)
    Symbol.uniq_id := 0
 

let () =
  List.iter create_func library_functions;
  (* HACK: Initialize uniq_id *)
  Symbol.uniq_id := 1

%}

%token T_eof 

%token<int> T_const 
%token<string> T_id 
%token<char> T_letter 
%token<string> T_string      

%token T_and T_bool T_char T_decl T_def T_else T_elsif T_end T_exit T_false 
  T_for T_head T_if T_int T_list T_mod T_new T_nil T_nilQ T_not T_or T_ref 
  T_return T_skip T_tail T_true T_plus T_minus T_mul T_div T_eq T_ins T_neq     
  T_less T_more T_leq T_meq T_lpar T_rpar T_lbra T_rbra T_comma T_colon T_dots
  T_assign

%left T_or
%left T_and
%left T_not             /*(*Just like unary +,- trick *)*/ 
%nonassoc T_eq T_neq T_more T_less T_leq T_meq
%right T_ins
%left T_plus T_minus
%left T_mul T_div T_mod
%left UPLUS UMINUS 

%start program
/*(* Non-Terminals *)*/
%type <unit> program
%type <sem_val> expr
%type <sem_val> call
%type <sem_val> stmt
%type <Symbol.entry> func_def
%%

program : 
     func_def T_eof 
      { 
        let start_pos = Parsing.rhs_start_pos 3 in
        let end_pos = Parsing.rhs_end_pos 3 in
        (* Check if main has formal parameters! NOT ALLOWED. *)
        match $1.entry_info with 
	| ENTRY_function f -> begin
	    if (List.length f.function_paramlist > 0) then	       
	      error "%a Main function parameters are not allowed" print_position
		(position_context start_pos end_pos)
	end
	| _ -> internal "%a Non-function before EOF." print_position
	      (position_context start_pos end_pos);
      }

/*(* Since we do not build an AST we have to break function definition rule in 
 * 3 because of what we call "NEXTQUAD-curse" that oblige us to work during 
 * summer. *)*/

/*(* In func_def we add the end quadruple: endu, f, -, - .
 * Every other quad (i.e. of local-defs and of function-body)
 * is already built. *)*/
func_def : 
     func_def_half compound_stmt T_end 
     {
       U.print ("func_def: end-of-fun " ^ (id_name $1.entry_id));
       let f, no = Stack.pop func_stack in
       match f.entry_info with
       | ENTRY_function inf -> 
           if (no = 0) && not(equalType inf.function_result TYPE_proc) then 
             let start_pos = Parsing.rhs_start_pos 3 in
             let end_pos = Parsing.rhs_end_pos 3 in 
             error "%a No 'return' statement in function's body" print_position
             (position_context start_pos end_pos);
	   else begin
             (* quads *)
             I.backpatch $2.l_next (nextQuad ());
             (* Remove useless quad: ret,-,-,- just before endu,I,-,- *)
             (*(* XXX: REMOVED! Serious bug in optimizations.... *)
	     let temp =  
               match !I.quad_list with
               | h :: t -> 
                   if (h.opor = I.O_Ret) then t
                   else (h :: t)
               | [] -> []
             in 
             quad_list := temp; (*decr Symbol.quadNext; decr quadNext no *)
	     *)
             add_quad (genQuad I.O_Endu (I.Entry $1) I.Empty I.Empty)
           end;
	   (* Store function's negative_offset for usage in Final. This could 
	    * be avoided if we generated code for quads and asm here (generation
	    * per Unit). As a matter of clarity we generate this code in Main.*)
	   let negofs = !currentScope.sco_negofs and
	       entry_list = !currentScope.sco_entries 
	   in
	   Final.func_offset_push ($1, negofs, entry_list);
	   (* Close scope at the end of function parsing *)
	   closeScope ();
           $1
       | _ -> fatal "Wrong entry in your stack, dude!"; $1
     }

/*(* In func_def_half we add start quadruple: unit, f, -, - .
 * Local-def quads must have been built. That's why we broke the rules at 
 * first.*)*/
func_def_half :  
   | T_def func_head T_dots  local_fun_defs 
   {
     U.print "func_def_half: end of locals";
     (* quads *)
     add_quad (genQuad I.O_Unit (I.Entry $2) I.Empty I.Empty);
     $2
   }
  
/*(* In func_head action we just do parameter semantics *)*/
func_head : 
   | T_id T_lpar formal_list T_rpar 
     {
       U.print ("func_head: procedure " ^ $1);
       let p = newFunction (id_make $1) true in 
       Stack.push (p, 0) func_stack;
       openScope ();
       let insert_param (ty, ids, pass) =
         List.iter (fun name -> ignore(newParameter (id_make name) ty pass p
         true)) ids in
       List.iter insert_param (List.rev $3);
       endFunctionHeader p TYPE_proc;
       p
     }
   | typee T_id T_lpar formal_list T_rpar 
     {
       U.print ("func_head: function " ^ $2);
       let p = newFunction (id_make $2) true in
       Stack.push (p, 0) func_stack;
       openScope ();
       let insert_param (ty, ids, pass) =
         List.iter (fun name -> ignore(newParameter (id_make name) ty pass p 
         true)) ids in
       List.iter insert_param (List.rev $4);
       endFunctionHeader p $1;
       p
     }

local_fun_defs : 
   | /*empty*/ {()}
   | local_fun_defs local_def {()}

local_def : 
   | func_def 
     { U.print "local_def: function definition" }
   | func_decl 
     { U.print "local_def: function declaration" }
   | var_def 
     { U.print "local_def: variable" }
  
var_def : 
     typee id_list 
     {
       List.iter (fun var -> ignore(newVariable (id_make var) $1 true)) $2
     }

func_decl :
     T_decl func_head 
     {
       U.print "func_decl: function declaration";
       forwardFunction $2;
       closeScope (); (* Declaring a forward function doesn't need to open a new
                         scope or put function in func_stack! *)
       ignore(Stack.pop func_stack);
     }

/*(* We break <formal_list> rule into two in order to accept functions that do
 * not need any parameters (and avoid classic shift/reduce error). *)*/
formal_list:
   | /*empty*/       
   { U.print "formal_list: empty"; [] }
   | formal_foo_list 
   { U.print "formal_list: non-empty"; $1 }

formal_foo_list : 
   | formal                         {[$1]}
   | formal_foo_list T_colon formal {$3 :: $1}

formal : 
   | typee id_list       
   { $1, List.rev $2, PASS_BY_VALUE }
   | T_ref typee id_list 
   { $2, List.rev $3, PASS_BY_REFERENCE }

id_list : 
   | T_id 
     {
       [$1]
     }
   | id_list T_comma T_id 
     {
       (*List.iter (printf "%s, ") $1;*)
       $3 :: $1
     } 

typee : 
   | T_int                         
     {
       U.print "typee: int";
       TYPE_int
     } 
   | T_bool 
     {
       U.print "typee: bool";
       TYPE_bool
     } 
   | T_char
     {
       U.print "typee: char";
       TYPE_char
     } 
   | typee T_lbra T_rbra           
     {
       U.print ("typee: array of " ^ (string_of_type $1));
       TYPE_array ($1, 0)
    } 
   | T_list T_lbra typee T_rbra   
     {
       U.print ("typee: list of " ^ (string_of_type $3));
       TYPE_list ($3, 0)
     } 

/*(*XXX: Empty stmt/function body is not allowed by Tony's EBNF *)*/
compound_stmt : 
   | stmt 
   {
     I.backpatch $1.l_next (nextQuad ());
     {def_sem with l_next = $1.l_next}
   }
   | compound_stmt foo_stmt stmt 
   {
     I.backpatch $1.l_next $2;
     I.backpatch $3.l_next (nextQuad ());
     {def_sem with l_next = $3.l_next}
   }

foo_stmt :
   /* empty */ {nextQuad ()}

stmt : 
   | simple {$1}
   | T_exit 
     {
      let f, no_of_exits = Stack.pop func_stack in
      match f.entry_info with
      | ENTRY_function inf -> 
        if (not (equalType inf.function_result TYPE_proc)) then begin
          let start_pos = Parsing.rhs_start_pos 3 in
          let end_pos = Parsing.rhs_end_pos 3 in 
          error "%a Non procedure object tries to 'exit'."
          print_position (position_context start_pos end_pos);
        end else 
          Stack.push (f, no_of_exits+1) func_stack;
          (* quads *)
          (*add_quad (genQuad I.O_Ret I.Empty I.Empty I.Empty);*)
          def_sem
      | _ -> 
          fatal "Wrong entry in your stack, dude!";
          def_sem
     }
   | T_return expr 
     {
      U.print "stmt: return";
      let f, no_of_rets = Stack.pop func_stack in
      match f.entry_info with
      | ENTRY_function inf -> 
        if (not (equalType inf.function_result $2.expr_type)) then 
          let start_pos = Parsing.rhs_start_pos 3 in
          let end_pos = Parsing.rhs_end_pos 3 in 
          error "%a Function should return <%s> but return value is of type <%s>"
          print_position (position_context start_pos end_pos)
          (string_of_type inf.function_result) (string_of_type $2.expr_type);
        else 
          Stack.push (f, no_of_rets+1) func_stack;
          (* quads *)
	  let rtype = $2.expr_type in
          add_quad (genQuad I.O_Assign $2.place I.Empty (I.Result rtype));
          add_quad (genQuad I.O_Ret I.Empty I.Empty I.Empty);
          $2
      | _ -> 
          fatal "Wrong entry in your stack, dude!";
          def_sem
     }
   /*(*We have to break <if_stmt> rule. Here {P_41} is executed. See
    * <if-head> rule, above.*)*/
   | if_stmt 
   {
     {def_sem with l_next = $1.l_next}
   }
   /*(*We have to break <for_stmt> too.*)*/
   | for_stmt 
   {
     {def_sem with l_next = $1.l_next}
   }

/*(*if <cond1> : {P_38} <compound1> 
* [elsif {P_38_1} <cond2> : {P_38_2} <compound2> {P_38_3}]
* [else : {P_39} <compound3> {P_40}] 
* end {P_41} *)*/
      
/*(* Backpatch(<cond1>.FALSE, <cond2>.quad_num) % (if_head, elsif_stmt).
 * Backpatch(<cond2>.FALSE, <compound3>.quad_num) % (elsif_stmt, else_stmt).
 * <if_stmt>.NEXT = <compound1>.NEXT @ [<jump1>.quad_num] 
 *       @ <compound2>.NEXT @ [<jump2>.quad_num] @ <compound3>.NEXT 
 * % (elsif_stmt, elsif_stmt, else_stmt, else_stmt) *)*/
if_stmt :
   | if_head compound_stmt else_stmt T_end /*(*{P_41}*)*/
   {
     U.print "if-else";
     let cond1_false = $1 and (else_next, compound3_num) = $3 in
     let res = {def_sem with l_next = else_next} in
     match compound3_num with
     | h :: _ -> 
         I.backpatch cond1_false h; res
     | [] -> 
         I.backpatch cond1_false (nextQuad ()); res (*Case of empty else*)
   }
   | if_head compound_stmt elsif_stmt else_stmt T_end /*(*{P_41}*)*/
   {
     U.print "if-elsif-else";
     let cond1_false = $1 
     and elsif_next, cond2_num_false = $3 
     and else_next, compound3_num = $4 in
     I.backpatch cond1_false (List.hd cond2_num_false);
     let res = {def_sem with l_next = elsif_next @ else_next @ $2.l_next} in
     match compound3_num with
     | h :: _ -> 
         I.backpatch (List.tl cond2_num_false) h; res
     | [] -> 
         {res with l_next = res.l_next @ (List.tl cond2_num_false)};
   }
     
/*(* Backpatch(<cond1>.TRUE, NEXTQUAD()) (<compound1>.quad_num) 
 * Return <cond1>.FALSE. *)*/
if_head :
     T_if cond T_dots /*(*{P_38}*)*/
     {
       U.print "if_head: backpatching cond of if";
       I.backpatch $2.l_true (I.nextQuad ());
       $2.l_false
     }

/*(* Return <jump2>.quad_num % (else_head).
 * Return <compound3>.NEXT. *)*/
else_stmt :
   | /*empty*/ 
   {
     ([], [])
   }
   | else_head compound_stmt /*(*{P_40}*)*/
   {
     let (jump2_num, compound3_num) = $1 in
     (jump2_num :: $2.l_next, [compound3_num])
   }

/*(* Return NEXTQUAD() (<jump2>.quad_num).
 * Add jump, -, -, *. 
 * Return NEXTQUAD() (<compound3>.quad_num). *)*/
else_head :
     T_else T_dots /*(*{P_39}*)*/
     {
       let jump2_num = nextQuad () in
       add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
       let compound3_num = nextQuad () in
       (jump2_num , compound3_num)
     }

/*(* Return elsif_head.
 * Return <compound2>.NEXT. *)*/
elsif_stmt :
     elsif_head compound_stmt /*(*{P_38_3}*)*/
     {
       match $1 with 
       | (jump1_num, cond2_num_false) -> (jump1_num :: $2.l_next, cond2_num_false)
       | _ -> internal "<elsif_head> in Elsif_stmt returns shit."; 
              ([],[]) (* never matched! *)
     }
   | elsif_stmt foo_stmt elsif_head compound_stmt 
     {
       match $1, $3 with 
       | (jumps_nexts, (cond_num :: cond_false)),
	 (jump1_num , (cond2_num :: cond2_false)) -> 
	   I.backpatch cond_false cond2_num;
	   (jumps_nexts @ (jump1_num :: $4.l_next), (cond_num :: cond2_false))
       | _ -> 
	   internal "Elsif_stmt returns shit."; 
           ([],[]) (* never matched! *)
     }

/*(* Return <cond2>.quad_num % (elsif_head_if).
 * Return <jump1>.quad_num % (elsif_head_if).
 * Backpatch(<cond2>.TRUE, NEXTQUAD()) (<stmt2>.quad_num) % (cond). 
 * Return <cond2>.FALSE. *)*/
elsif_head :
     elsif_head_if cond T_dots /*(*{P_38_2}*)*/
     {
       let jump1_num, cond2_num = $1 in
       I.backpatch $2.l_true (nextQuad ());
       (jump1_num , cond2_num :: $2.l_false)
     }

/*(* Return NEXTQUAD() (<jump1>.quad_num).
 * Add jump, -, -, *.  
 * Return NEXTQUAD() (<cond2>.quad_num). *)*/
elsif_head_if :
     T_elsif /*(*{P_38_1}*)*/
     {
      let q1 = nextQuad () in
      add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
      (q1 , nextQuad ())
     }

/*(*--------- end of IF ---------*)*/        
        
/*(* for <simple_list> ; {P_42} <cond> ; {P_43} <simple_list> : {P_44}
 * <stmt>+ {P_45}
 * end  *)*/

/*(* Add jump, -, -, <simple2>.quad_num 
 * <for_stmt>.NEXT = <cond>.FALSE *)*/
for_stmt :
     for_head compound_stmt T_end /*(*{P_45}*)*/
     {
       let cond_false, simple2_quad = $1 in
       add_quad (genQuad I.O_Jump I.Empty I.Empty (I.Label simple2_quad));
       {def_sem with l_next = cond_false}
     }

/*(* Add jump, -, -, <cond>.quad_num % (for_head_fst).
 * Backpatch(<cond>.TRUE, NEXTQUAD()) 
 * % (for_head_snd, <for_body>.quad_num) 
 * Return <cond>.FALSE and <simple2>.quad_num (by for_head_snd). *)*/
for_head :
     for_head_fst for_head_snd simple_list T_dots /*(*{P_44}*)*/
     {
       let cond_quad = $1 and
       cond_true, cond_false, simple2_quad = $2 and
       body_fst_quad = nextQuad () + 1 in
       add_quad (genQuad I.O_Jump I.Empty I.Empty (I.Label cond_quad));
       I.backpatch cond_true body_fst_quad;
       (cond_false, simple2_quad)
     }

/*(* Return NEXTQUAD() (<cond>.quad_num). *)*/
for_head_fst : 
     T_for simple_list T_colon /*(*{P_42}*)*/
     {
       let cond_quad = nextQuad () in
       cond_quad
     }

/*(* Return <cond>.TRUE, <cond.FALSE>
 * Return NEXTQUAD() (<simple2>.quad_num). *)*/
for_head_snd :
     cond T_colon /*(*{P_43}*)*/
     {
       let simple2_quad = nextQuad () in
       ($1.l_true, $1.l_false, simple2_quad)
     }

/*(*--------- end of FOR ---------*)*/        

simple : 
   | T_skip                
   {
     U.print "simple stmt: skip";
     def_sem
   }
   | l_value T_assign expr 
     {
       U.print "simple stmt: assign statement";
       if ($1.value <> Lval) then (* string constant *) begin
         let start_pos = Parsing.rhs_start_pos 3 in
         let end_pos = Parsing.rhs_end_pos 3 in 
         error "%a Cannot assign to something that is not an L-value!"
         print_position (position_context start_pos end_pos)
       end;
       if not (equalType $1.expr_type $3.expr_type) then begin
         let start_pos = Parsing.rhs_start_pos 3 in
         let end_pos = Parsing.rhs_end_pos 3 in 
         error "%a Incompatible types in assignment! You cannot assign an \
         expression of type '%s' to an expression of type '%s'." 
         print_position (position_context start_pos end_pos) (string_of_type
         $3.expr_type) (string_of_type $1.expr_type);
       end;
       (* quads *)
       add_quad (genQuad I.O_Assign $3.place I.Empty $1.place);
       def_sem
     }
   | call                  
     {
       U.print "simple stmt: call";
       {def_sem with place = $1.place}
       (* XXX: Should we allow non-procedure call to be executed as simple
       	* stmt? *)
     } 

simple_list :
   | simple {()}
   | simple_list T_comma simple {()}

func_actual_params :
   | /*empty*/      {[]}
   | param_foo_list  {$1}

param_foo_list :
   | expr                       {[$1]}
   /*| expr 
     {
     let isTempo (a:I.operand_t) = 
       match a with
       | I.Entry e ->
	   begin
	   match e.entry_info with
	   | ENTRY_temporary _ -> true
	   | _ -> false
	   end
       | _ -> false
     in
     (*XXX: Fix f(x, inc(x), x) sideffect bug! *)
     if not(isTempo $1.place) then
       let w = newTemporary $1.expr_type in
       add_quad (genQuad I.O_Assign $1.place I.Empty (I.Entry w));
       [{$1 with place = I.Entry w}]
     else
       [$1]
     }*/
   | param_foo_list T_comma expr {
     (*let w = newTemporary $3.expr_type in
     * add_quad (genQuad I.O_Assign $3.place I.Empty (I.Entry w));
     * let res = {$3 with place = I.Entry w} in *)
     $3 :: $1
}

call :
     T_id T_lpar func_actual_params T_rpar 
     { 
       U.print ("call: " ^ $1);
       let start_pos = Parsing.rhs_start_pos 3
       and end_pos = Parsing.rhs_end_pos 3
       and f = lookupEntry (id_make $1) LOOKUP_ALL_SCOPES true
       and param_list = (List.rev $3) in
       match f.entry_info with
       | ENTRY_function inf ->  
           (* Check compatibility in arity. *)
           let l1 = List.length inf.function_paramlist
           and l2 = List.length param_list in
           if (l1<>l2) then begin
             error "%a Function '%s' needs %d parameters but is here called with \
             %d parameters!"print_position (position_context start_pos end_pos)
             $1 l1 l2
           end;
           (* Check type/pass_mode compatibility. *)
           let type_compatible (x:entry) (y:sem_val) = 
	     begin
               match x.entry_info with
               | ENTRY_parameter p -> 
                   if (equalType p.parameter_type y.expr_type) then begin
                   (* pass_mode compatibility *)
                     match p.parameter_mode with
                     | PASS_BY_REFERENCE -> 
			 if (y.value <> Lval) then 
                         error "%a Function '%s': Actual parameter should be an \
                         L-Value of type '%s' (call by reference)." print_position 
                         (position_context start_pos end_pos) $1 (string_of_type
                         p.parameter_type)
                       else  
                         (* quads *)
			 (*let w = newTemporary y.expr_type in 
			  *add_quad (genQuad I.O_Assign y.place I.Empty (I.Entry w));*)
                         add_quad (genQuad I.O_Par y.place (I.Pass R) I.Empty)
                   | PASS_BY_VALUE -> 
                       (* quads *)
                       (*let w = newTemporary y.expr_type in 
                       add_quad (genQuad I.O_Assign y.place I.Empty (I.Entry w));*)
                       add_quad (genQuad I.O_Par y.place (I.Pass V) I.Empty)
		   end
                   else
                   error "%a Function '%s': Typical parameter '%s' is not \
                   compatible in type with actual parameter!" print_position 
                   (position_context start_pos end_pos) $1 (id_name 
                   x.entry_id)
               | _ -> 
                   error  "%a Function '%s': Checking compatibility cannot be \
		     completed! '%s' is not an ENTRY_parameter!" print_position 
                     (position_context start_pos end_pos)$1 (id_name x.entry_id)
	     end
           in
	   List.iter2 type_compatible inf.function_paramlist param_list;
           let result = {def_sem with value = Rval; expr_type = 
             inf.function_result;} 
           in
           (* quads *)
           if (not (equalType inf.function_result TYPE_proc)) then begin
             let w = newTemporary inf.function_result in
             add_quad (genQuad I.O_Par (I.Entry w) (I.Pass RET) I.Empty);
             result.place <- I.Entry w
           end;
           add_quad (genQuad I.O_Call I.Empty I.Empty (I.Entry f));
           result
       | ENTRY_variable _ ->
           fatal "Symbol '%s' is a variable, but is here used as a function." $1;
           def_sem
       | ENTRY_parameter _ ->
           fatal "Symbol '%s' is a function parameter, but is here used as a \
           function." $1;
           def_sem
       | _ ->
           internal "Strange expression disguises herself as a function call";
           def_sem
     }
/*(* ------------ end of call ------------ *)*/

l_value : 
   | T_id 
     {
       U.print ("l_value: identifier " ^ $1);
       let l = lookupEntry (id_make $1) LOOKUP_ALL_SCOPES true in
       match l.entry_info with 
       | ENTRY_parameter p -> 
           {def_sem with value = Lval; expr_type = p.parameter_type; 
           place = I.Entry l}
       | ENTRY_variable v -> 
           {def_sem with value = Lval; expr_type = v.variable_type; 
           place = I.Entry l}
       | _ -> begin 
         let start_pos = Parsing.rhs_start_pos 3 in
         let end_pos = Parsing.rhs_end_pos 3 in 
         error "%a Not a valid l-value: %s" print_position (position_context
         start_pos end_pos) $1; 
         def_sem 
         end
     }
   | T_string 
     {
       U.print ("l_value: string " ^ $1);
       let len = String.length $1 in 
       {def_sem with expr_type = TYPE_array (TYPE_char, len+1); place= I.Str $1;
       value = Lval}  (* T_string Lval "abc"[0] -> "a"! *)
     }
   | l_value T_lbra expr T_rbra 
     {
       U.print ("l_value: array");
       let start_pos = Parsing.rhs_start_pos 3 
       and end_pos = Parsing.rhs_end_pos 3 in 
       if (not (equalType TYPE_int $3.expr_type)) then begin
         error "%a Array/List index should be of type: int." print_position
         (position_context start_pos end_pos); def_sem
       end else begin
         let ty = U.extract_type $1.expr_type in 
         let w = newTemporary TYPE_int in
         (* quads *)
         add_quad (genQuad I.O_Array $1.place $3.place (I.Entry w));
         {def_sem with value = Lval; expr_type = ty; place = I.Pointer (w, ty)}
       end
     }
   | call 
     {
       U.print "l_value: call";
       {$1 with value = Lval}
     } 

expr : 
   | l_value 
     {
      U.print "expr: <l_value>";
      $1
     }
   | T_true 
     { 
       {def_sem with expr_type = TYPE_bool; place = I.Boole true; l_true =
         [nextQuad ()]}
     }
   | T_false
     { 
       {def_sem with expr_type = TYPE_bool; place = I.Boole false; l_false =
         [nextQuad ()]}
     }
   | T_const 
     {
      U.print ("expr: <int-const> " ^ (string_of_int $1));
      {def_sem with expr_type = TYPE_int; place = I.Int $1}
     }
   | T_letter 
     {
       U.print ("expr: <char-const> " ^ (Char.escaped $1));
       {def_sem with expr_type = TYPE_char; place = I.Char $1}
     } 
   | T_nil 
     {
       U.print "expr: nil";
       {def_sem with expr_type = TYPE_list(TYPE_none,0); place =
         I.Nil}
     }
   | T_lpar expr T_rpar {$2}
   /*(*XXX: <cond> is not allowed as an expression! i.e bool b1 := true or \
	false is a syntax error. *)*/
   | expr_binary 
     {
      let t1, t2, op = $1 in
      if (not (equalType t1.expr_type TYPE_int) || 
          not (equalType t2.expr_type TYPE_int)) then begin
            let start_pos = Parsing.rhs_start_pos 3 
            and end_pos = Parsing.rhs_end_pos 3 in 
            error "%a Type mismatch: Operator '%s' has type int '%s' int."
            print_position (position_context start_pos end_pos) 
            (str_of_operator op) ("*"); 
            def_sem 
      end else begin
         (* quads *)
	 match (t1.place, t2.place) with 
	 | (I.Int x, I.Int y) -> begin
	 (* Constant folding optimization *)
	     let func = binop_of_operator op in
	     {def_sem with expr_type = TYPE_int; place = I.Int (func x y)}
	 end;
	 | _ -> begin
             let w = newTemporary TYPE_int in 
             add_quad (genQuad op t1.place t2.place (I.Entry w));
             {def_sem with expr_type = TYPE_int; place = (I.Entry w)} 
	 end;
      end
     }
   | expr_unop
     {
      let t1, sign = $1 in
      if (not (equalType t1.expr_type TYPE_int)) then begin
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a Type mismatch: Unary operator '%s' is applicable only on expressions \
        of type int." print_position (position_context start_pos end_pos) 
        (str_of_operator sign); 
        def_sem 
      end else begin
        (* quads *)
        if (sign = I.O_Minus) then 
	  match t1.place with 
	  | I.Int x -> 
	      (* Constant folding optimization *)
	      {def_sem with expr_type = TYPE_int; place = I.Int (-x)}
	  | _ -> begin
              let w = newTemporary TYPE_int in
              let q = genQuad sign t1.place I.Empty (I.Entry w) in 
              add_quad q;
              {def_sem with expr_type = TYPE_int; place = I.Entry w} 
	  end;
        else
          {def_sem with expr_type = TYPE_int; place = t1.place}
      end
     }
   | T_new typee T_lbra expr T_rbra  
     {
      let start_pos = Parsing.rhs_start_pos 3 
      and end_pos = Parsing.rhs_end_pos 3 
      in 
      if (not (equalType $4.expr_type TYPE_int)) then begin
        error "%a Size of array must be an expression of type: int."
        print_position (position_context start_pos end_pos);
        def_sem
      end else
        (* quads *)
        let size = sizeOfType $2 and temp = newTemporary TYPE_int in
	if U.is_v $2 then begin
	  (* parameter in newarrayv is size in BYTES *)
          add_quad (genQuad I.O_Mult $4.place (I.Int size) (I.Entry temp));
          add_quad (genQuad I.O_Par (I.Entry temp) (I.Pass V) I.Empty);
          let w = newTemporary TYPE_int in 
          add_quad (genQuad I.O_Par  (I.Entry w) (I.Pass RET) I.Empty); 
          add_quad (genQuad I.O_Call I.Empty I.Empty I.Newv);
	  (* GC: record newarrayv for making appropriate register_call_table calls*)
	  (*Final.push_extern_call_table "_newarrayv";*)
          {def_sem with value = Lval; expr_type = TYPE_array ($2,size); place =
            I.Entry w} (* I.Pointer (w, $2) *)
	end else (* is_pointer *) begin
	  (* parameter in newarrayp is size in WORDS *)
	  add_quad (genQuad I.O_Assign $4.place I.Empty (I.Entry temp));	  
          add_quad (genQuad I.O_Par (I.Entry temp) (I.Pass V) I.Empty);
          let w = newTemporary TYPE_int in 
          add_quad (genQuad I.O_Par  (I.Entry w) (I.Pass RET) I.Empty); 
          add_quad (genQuad I.O_Call I.Empty I.Empty I.Newp);
	  (* GC: record newarrayp for making appropriate register_call_table calls*)
	  (*Final.push_extern_call_table "_newarrayp";*)
          {def_sem with value = Lval; expr_type = TYPE_array ($2,size); place =
            I.Entry w} (* I.Pointer (w, $2) *)
	end;
     }
   | expr T_ins expr 
     { (* List constructor *)
      U.print "stmt: expr # expr";
      let start_pos = Parsing.rhs_start_pos 3 
      and end_pos = Parsing.rhs_end_pos 3
      in
      let ty1 = $1.expr_type and ty2 = $3.expr_type in
      match ty2 with
      | TYPE_list (tp, sz) -> 
        (* Empty list is compatible with any type for a 'cons' operation. *)
        if (equalType ty1 tp || sz = 0) then begin
          (* quads *)
	  if U.needs_extend ty1 then begin
	    let y = newTemporary TYPE_int in (* extend head to word *)
	    add_quad (genQuad I.O_Par $1.place (I.Pass V) I.Empty); 
	    add_quad (genQuad I.O_Par (I.Entry y) (I.Pass RET) I.Empty);
	    add_quad (genQuad I.O_Call I.Empty I.Empty I.Extend);
	    (* pass extended var *)
	    add_quad (genQuad I.O_Par (I.Entry y) (I.Pass V) I.Empty);
	  end else 
	    add_quad (genQuad I.O_Par $1.place (I.Pass V) I.Empty);
          add_quad (genQuad I.O_Par $3.place (I.Pass V) I.Empty);
          let w = newTemporary (TYPE_list (tp, sz + 1)) in
          add_quad (genQuad I.O_Par (I.Entry w) (I.Pass RET) I.Empty);
	  (* Make appropriate call *)
	  if U.is_v ty1 then (* Head is value *) begin
            add_quad (genQuad I.O_Call I.Empty I.Empty I.Consv);
            (* GC: record consv for making appropriate register_call_table calls*)
	    Final.push_extern_call_table "_consv"
	  end else (* Head is pointer *) begin
            add_quad (genQuad I.O_Call I.Empty I.Empty I.Consp);
            (* GC: record consp for making appropriate register_call_table calls*)
	    Final.push_extern_call_table "_consp"
          end;
	  {def_sem with value = Lval; expr_type = TYPE_list (tp, sz+1); place = I.Entry w}
        end else begin
          error "%a Cannot construct a list by an element of type '%s' (head) and \
          a list of type '%s' (tail)." print_position (position_context start_pos
          end_pos) (string_of_type ty1) (string_of_type ty2);
          def_sem end
      | _                 -> 
        error "%a Second argument should have type: list of %s." 
          print_position (position_context start_pos end_pos) (string_of_type
          ty1);
        def_sem
     }
   | T_head T_lpar expr T_rpar 
     {
      U.print "expr: 'head'";
      match $3.expr_type with
      | TYPE_list (tp,sz) -> 
        (* quads *)
        let w = newTemporary TYPE_int and res = {def_sem with expr_type = tp} 
	in  (* head always returns word! *)
        add_quad (genQuad I.O_Par $3.place (I.Pass V) I.Empty);
        add_quad (genQuad I.O_Par (I.Entry w) (I.Pass RET) I.Empty);
        add_quad (genQuad I.O_Call I.Empty I.Empty I.Head);
        res.place <- I.Entry w;
	if U.needs_extend tp then begin (* shrink to byte *)
	  let y = newTemporary tp in
	  add_quad (genQuad I.O_Par (I.Entry w) (I.Pass V) I.Empty); 
	  add_quad (genQuad I.O_Par (I.Entry y) (I.Pass RET) I.Empty);
	  add_quad (genQuad I.O_Call I.Empty I.Empty I.Shrink);
	  res.place <- I.Entry y
	end;
	res
          (*if (sz>0) then {def_sem with expr_type = tp}
          else begin
            error "'head' operator cannot be applied to empty lists!";
            def_sem
          end*)
      | _                 -> 
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a 'head' operator can only be applied to arguments of type list[t]."
        print_position (position_context start_pos end_pos);
        def_sem
     }
   | T_tail T_lpar expr T_rpar 
     {
      U.print "expr: 'tail'";
      match $3.expr_type with
      | TYPE_list (tp,sz) ->
        (* quads *)
        let w = newTemporary (TYPE_list (tp, sz-1)) in
        add_quad (genQuad I.O_Par $3.place (I.Pass V) I.Empty);
        add_quad (genQuad I.O_Par (I.Entry w) (I.Pass RET) I.Empty);
        add_quad (genQuad I.O_Call I.Empty I.Empty I.Tail);
        {def_sem with value = Lval; expr_type = $3.expr_type; place = I.Entry w}
          (*if (sz>0) then {def_sem with expr_type = $3.expr_type}
             else begin
             error "'head' operator cannot be applied to empty lists!";
             def_sem
          end*)
      | _                 -> 
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a 'tail' operator can only be applied to arguments of type list[t]."
        print_position (position_context start_pos end_pos);
        def_sem
     }
     /*(*
     | T_nilQ T_lpar expr T_rpar 
     { 
     match $3.expr_type with
     | TYPE_list (tp,sz) -> 
        (* quads *)
	 let b = newTemporary TYPE_bool in 
	 add_quad (genQuad I.O_Par $3.place (I.Pass V) I.Empty);
	 add_quad (genQuad I.O_Par (I.Entry b) (I.Pass RET) I.Empty);
         add_quad (genQuad I.O_Call I.Empty I.Empty I.NilQ);

         {def_sem with expr_type = TYPE_bool; place = I.Entry b} 
     | _  ->  
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a nil? operator can only be applied to arguments of type list[t]."
          print_position (position_context start_pos end_pos);
        def_sem 
     }
     *)*/
  

/*(* We have to break and/or rules because of "short-circuit" 
 * application *)*/
cond :
   | cond_minimal 
   { (* quads *)
    (*let quad_true = I.nextQuad () in
    add_quad (genQuad I.O_If_jump $1.place I.Empty I.Backpatch);
    let quad_false = I.nextQuad () in
    add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
    {$1 with l_true = [quad_true]; l_false = [quad_false]}*)
    $1
   }
   | cond_and_half cond_minimal 
     { 
      if (not (equalType $1.expr_type TYPE_bool) || 
          not (equalType $2.expr_type TYPE_bool)) then begin
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a Type mismatch: Operator 'and' has type bool '%s' bool. \n\
        %s <> %s \n"  
          print_position (position_context start_pos end_pos) ("*")
          (string_of_type $1.expr_type)          (string_of_type $2.expr_type);
        def_sem 
      end else 
        {def_sem with expr_type = TYPE_bool; l_true = $2.l_true; l_false =
          $1.l_false @ $2.l_false}
     }
   | cond_or_half cond_minimal
     { 
      if (not (equalType $1.expr_type TYPE_bool) || 
          not (equalType $2.expr_type TYPE_bool)) then begin
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a Type mismatch: Operator 'or' has type bool '%s' bool."
          print_position (position_context start_pos end_pos) ("*");
        def_sem 
      end else 
        {def_sem with expr_type = TYPE_bool; l_false = $2.l_false; l_true =
          $1.l_true @ $2.l_true}
     }

cond_and_half :
     cond T_and 
     { (* quads *)
       I.backpatch $1.l_true (I.nextQuad ());
       $1
     }

cond_or_half :
     cond T_or  
     { (* quads *)
       I.backpatch $1.l_false (I.nextQuad ());
       $1
     }

cond_minimal : 
   | T_not cond 
     {
      if (not (equalType $2.expr_type TYPE_bool)) then begin
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a Unary operand '%s' is appliable only on expressions of type \
        bool." print_position (position_context start_pos end_pos) "not";
        def_sem 
      end else 
        {def_sem with expr_type = TYPE_bool; l_true = $2.l_false; l_false = 
          $2.l_true}
     }
   | T_lpar cond_minimal T_rpar {$2 (*XXX: 1 shift/reduce from this!*)}
   | expr_relop 
     {
      let t1, t2, op = $1 in
      if not(equalType t1.expr_type t2.expr_type) then begin
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a Type mismatch: Cannot use operator '%s' with type %s '%s' %s."
        print_position (position_context start_pos end_pos) (I.str_of_operator 
        op) (string_of_type t1.expr_type) ("*") (string_of_type t2.expr_type); 
        def_sem 
      end else begin
        (* quads *)
	match (t1.place, t2.place) with
	| (I.Boole b1, I.Boole b2) -> begin
	    (* Constant folding optimization *)
	    let func = relop_of_operator op in
	    let eval = (func b1 b2) in (* Evaluate relop *)
            let quad_true = nextQuad () in
            add_quad (genQuad I.O_If_jump (I.Boole eval) I.Empty I.Backpatch);
            let quad_false = nextQuad () in
            add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
	    {def_sem with expr_type = TYPE_bool; place = I.Boole true; l_true = [quad_true]; 
	      l_false = [quad_false]}
	end
	| (I.Int x1, I.Int x2) -> begin
	    let func = relop_of_operator op in
	    let eval = (func x1 x2) in (* Evaluate relop *)
            let quad_true = nextQuad () in
            add_quad (genQuad I.O_If_jump (I.Boole eval) I.Empty I.Backpatch);
            let quad_false = nextQuad () in
            add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
	    {def_sem with expr_type = TYPE_bool; place = I.Boole true; l_true = [quad_true]; 
	      l_false = [quad_false]}
	end
	| _ -> begin
            let quad_true = nextQuad () in
            add_quad (genQuad op t1.place t2.place I.Backpatch);
            let quad_false = nextQuad () in
            add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
            {def_sem with expr_type = TYPE_bool; l_true = [quad_true]; l_false =
              [quad_false]}
	end
      end
     }
   | expr
     {
      if not(equalType $1.expr_type TYPE_bool) then begin
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a Expression should be of type 'bool' when used as a boolean \
        condition." print_position (position_context start_pos end_pos);
        def_sem 
      end else begin
      (* quads *)
        U.print "cond_minimal: call of type BOOL";
        let q1 = nextQuad () in
        add_quad (genQuad I.O_If_jump $1.place I.Empty I.Backpatch);
        let q2 = nextQuad () in
        add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
        {def_sem with expr_type = $1.expr_type; l_true = [q1]; l_false = [q2]}
      end
     }
   | T_nilQ T_lpar expr T_rpar 
       { 
     match $3.expr_type with
     | TYPE_list (tp,sz) -> 
        (* quads *)
        let quad_true = nextQuad () in
        add_quad (genQuad I.O_Equal $3.place I.Nil I.Backpatch);
        let quad_false = nextQuad () in
        add_quad (genQuad I.O_Jump I.Empty I.Empty I.Backpatch);
        {def_sem with expr_type = TYPE_bool; l_true = [quad_true]; l_false =
          [quad_false]}
     | _  ->  
        let start_pos = Parsing.rhs_start_pos 3 
        and end_pos = Parsing.rhs_end_pos 3 in 
        error "%a nil? operator can only be applied to arguments of type list[t]."
          print_position (position_context start_pos end_pos);
        def_sem 
} 

expr_relop :
   | expr T_eq expr      {$1, $3, I.O_Equal}
   | expr T_neq expr     {$1, $3, I.O_Nequal}
   | expr T_less expr    {$1, $3, I.O_Less}
   | expr T_more expr    {$1, $3, I.O_Greater}
   | expr T_leq expr     {$1, $3, I.O_Lequal}
   | expr T_meq expr     {$1, $3, I.O_Gequal}

expr_unop:
   | T_plus expr %prec UPLUS    {$2, I.O_Plus}
   | T_minus expr %prec UMINUS  {$2, I.O_Minus}

expr_binary :
   | expr T_plus expr    {$1, $3, I.O_Plus}
   | expr T_minus expr   {$1, $3, I.O_Minus}
   | expr T_mul expr     {$1, $3, I.O_Mult}
   | expr T_div expr     {$1, $3, I.O_Div}
   | expr T_mod expr     {$1, $3, I.O_Mod}

%%
